from pyomo.environ import *

m = AbstractModel('A_S_R')

# Sets
m.I = Set() # index of products (i)
m.J = Set() # index of machines(j)
m.T = Set() # index of periods(t)
m.R = Set() # index of raw material(r)

# Data
m.d = Param(m.I, m.T)
m.ca = Param(m.J)
m.req = Param(m.I, m.J)
m.cp = Param(m.I, m.J)
m.cc = Param( m.J)
m.st = Param(m.I, m.J)
m.ur = Param(m.R, m.I)
m.InvIp = Param(m.I)
m.InvIr = Param(m.R)
m.chp = Param(m.I)
m.cd = Param(m.I)
m.sm = Param(m.J, m.T)
m.pb = Param(m.I, m.J, m.T)
m.P = Param(m.I, m.J, m.T)
M =999

# Decision variables
m.X = Var(m.I, m.J, m.T,  domain=Binary, initialize=0)
m.QP = Var(m.I, m.J, m.T, domain=NonNegativeIntegers, initialize=0)
m.BO = Var(m.I, m.T,      domain=NonNegativeIntegers, initialize=0)
m.STT = Var(m.J, m.T,     domain=NonNegativeIntegers, initialize=0)
m.InvFP = Var(m.I, m.T,   domain=NonNegativeIntegers, initialize=0)
m.InvFR = Var(m.R, m.T,   domain=NonNegativeIntegers, initialize=0)
m.XR = Var(m.I, m.J, m.T, domain=Binary, initialize=0)
m.Z = Var(m.J, m.T,       domain=Binary, initialize=0)
m.STMAX = Var(m.J, m.T,   domain=NonNegativeIntegers, initialize=0 )
m.CFT = Var(m.J, m.T,     domain=NonNegativeIntegers, initialize=0)


# EXPRESIONS
def CTPR_rule(m):
    return sum(m.cp[i, j] * m.QP[i, j, t] for i in m.I for j in m.J for t in m.T)
m.CTPR = Expression(rule=CTPR_rule)

def CTD_rule(m):
    return sum(m.cd[i] * m.BO[i, t] for i in m.I for t in m.T)
m.CTD = Expression(rule=CTD_rule)

def CTC_rule(m):
    return sum(m.cc[j] * m.CFT[j, t] for j in m.J for t in m.T)
m.CTC = Expression(rule=CTC_rule)

def TCINVP_rule(m):
    return sum(m.chp[i] * m.InvFP[i, t] for i in m.I for t in m.T)
m.TCINVP = Expression(rule = TCINVP_rule)

# OBJECTIVE
def obj_rule(m):
    return m.CTPR + m.CTD + m.CTC + m.TCINVP
m.obj = Objective(rule=obj_rule, sense=minimize)

# CONSTRAINTS
def R2a_rule(m, i, t):
    if not ((t == 1)):
        return Constraint.Skip
    else:
        return m.InvFP[i,t] == m.InvIp[i] + sum(m.QP[i, j, t] for j in m.J) - m.d[i, t] + sum(m.pb[i, j, t]for j in m.J) + m.BO[i, t]
m.R2a = Constraint(m.I, m.T, rule=R2a_rule)

def R2b_rule(m, i, t):
    if not ((t > 1)):
        return Constraint.Skip
    return m.InvFP[i,t] == m.InvFP[i,t-1] + sum(m.QP[i, j, t] for j in m.J) - m.d[i, t] + sum(m.pb[i, j, t]for j in m.J) + m.BO[i, t] - m.BO[i, t - 1]
m.R2b = Constraint(m.I, m.T, rule=R2b_rule)


def R3a_rule(m, r, t):
    if not ((t == 1)):
        return Constraint.Skip
    return m.InvFR[r,t] == m.InvIr[r] - sum(m.ur[r, i] * m.QP[i, j,t] for i in m.I for j in m.J)
m.R3a = Constraint(m.R, m.T, rule=R3a_rule)

def R3b_rule(m, r, t):
    if not ((t > 1)):
        return Constraint.Skip
    return m.InvFR[r,t] == m.InvFR[r,t-1] - sum(m.ur[r, i] * m.QP[i, j,t] for i in m.I for j in m.J)
m.R3b = Constraint(m.R, m.T, rule=R3b_rule)

def R4_rule(m,r,t):
    return sum(m.ur[r, i] * m.QP[i, j, t] for i in m.I for j in m.J) <= m.InvIr[r]
m.R4 = Constraint(m.R,m.T, rule=R4_rule)


def R5_rule(m, i):
    return sum(m.QP[i,j, t] for j in m.J for t in m.T) + m.InvIp[i] == sum(m.d[i, t] for t in m.T) - sum(m.pb[i, j, t] for j in m.J for t in m.T)
m.R5 = Constraint(m.I, rule = R5_rule)


def R6a_rule(m, j, t):
    if m.sm[j, t] == 1:
        return sum(m.X[i, j, t] for i in m.I) == 0
    else:
        return Constraint.Skip
m.R6a = Constraint(m.J,m.T, rule=R6a_rule)

def R6b_rule(m, j, t):
    if m.sm[j, t] == 1:
        return sum(m.req[i, j] * m.QP[i, j, t] for i in m.I) + m.STT[j, t] <= m.ca[j]
    else:
        return Constraint.Skip
m.R6b = Constraint(m.J,m.T, rule=R6b_rule)

def R7a_rule(m,i, j,t):
    if t == 1:
        return m.XR[i,j, t] == 0
    else:
        return Constraint.Skip
m.R7a = Constraint( m.I,m.J,m.T, rule=R7a_rule)

def R7b_rule(m, i, j, t):
    if t > 1:
        return m.XR[i, j, t] <= m.X[i, j, t]
    else:
        return Constraint.Skip
m.R7b = Constraint(m.I, m.J, m.T, rule=R7b_rule)

def R8_rule(m, i, j, t):
    if t > 1:
        return m.XR[i, j, t] <= m.X[i, j, t-1]
    else:
        return Constraint.Skip
m.R8 = Constraint(m.I, m.J, m.T, rule=R8_rule)

def R9_rule(m, i, j, t):
    if t > 1:
        return m.XR[i, j, t] >= m.X[i, j, t] + m.X[i, j, t-1] - 1
    else:
        return Constraint.Skip
m.R9 = Constraint(m.I, m.J, m.T, rule=R9_rule)


def R10_rule(m, j,t):
    if t == 1:
        return m.STT[j, t] == sum(m.st[i, j] * m.X[i, j, t] for i in m.I)
    else:
        return Constraint.Skip
m.R10 = Constraint(m.J,m.T, rule=R10_rule)

def R11_rule(m, j, t):
    if t > 1:
        return m.STT[j, t] >= sum(m.st[i, j] * m.X[i, j, t] for i in m.I) - m.STMAX[j, t]
    else:
        return Constraint.Skip
m.R11 = Constraint(m.J, m.T, rule=R11_rule)


def R12_rule(m, j, t):
    if t > 1:
        return m.STT[j, t] <= sum(m.st[i, j] * m.X[i, j, t] for i in m.I) - m.STMAX[j, t]
    else:
        return Constraint.Skip
m.R12 = Constraint(m.J, m.T, rule=R12_rule)


def R13_rule(m,i,j,t):
    if t > 1:
        return m.STMAX[j,t] >= m.st[i, j] * m.XR[i, j, t]
    else:
        return Constraint.Skip
m.R13 = Constraint(m.I,m.J,m.T, rule=R13_rule)

def R14_rule(m, j, t):
    if t > 1:
        return m.STMAX[j, t] <= sum(m.st[i, j] * m.XR[i, j, t] for i in m.I)
    else:
        return Constraint.Skip
m.R14 = Constraint(m.J, m.T, rule=R14_rule)

def R15_rule(m, j,t):
    if t == 1:
        return m.CFT[j, t] == sum(m.X[i, j, t] for i in m.I)
    else:
        return Constraint.Skip
m.R15 = Constraint( m.J,m.T, rule=R15_rule)

def R16_rule(m, j,t):
    if t > 1:
        return m.CFT[j, t] == sum(m.X[i, j, t] for i in m.I) - m.Z[j, t]
    else:
        return Constraint.Skip
m.R16 = Constraint( m.J,m.T, rule=R16_rule)


#Bounds
def bound_1_rule(m, i, j,t):
    return m.QP[i, j, t] <= M * m.X[i, j, t]
m.bound_1 = Constraint(m.I, m.J,m.T, rule=bound_1_rule)
def bound_2_rule(m, i, j,t):
    return m.X[i, j, t] <= M * m.QP[i, j,t]
m.bound_2 = Constraint(m.I, m.J,m.T, rule=bound_2_rule)
def bound_3_rule(m,j,t):
    return m.Z[j, t] <= M * sum(m.XR[i,j,t] for i in m.I)
m.bound_3 = Constraint( m.J,m.T, rule=bound_3_rule)

