# Problema de la Programación de Producción en Talleres Sujetos a Eventos Inesperados

## Descripción del Problema

Este Trabajo Fin de Máster propone un enfoque basado en un Modelo de Programación Lineal Mixta (MILP) para abordar el problema de programación de producción en talleres de trabajo sujetos a la aparición de eventos inesperados, con el objetivo de minimizar los costos totales asociados a la producción y a los cambios de programación. En una primera etapa se calcula un plan de producción inicial con la asignación óptima de lotes de producción a máquinas, considerando las operaciones requeridas. En sucesivas etapas se ejecuta la reprogramación dinámica del plan de producción inicial en respuesta a eventos inesperados, como averías de maquinaria o escasez de materias primas. La reprogramación se adapta al estado actual del taller de producción y se ejecuta con el objetivo de encontrar una nueva solución óptima o casi óptima que pueda adaptarse rápidamente a cambios en las condiciones del entorno de producción, con el fin de minimizar cambios significativos respecto al plan de producción inicial.


## Visión General del Modelo

### Índices

- `i`: Índice de productos.
- `j`: Índice de máquinas.
- `r`: Índice de primas.
- `t`: Índice de períodos de tiempo.

### Parámetros

- `d_it`: Demanda del producto `i` durante el período `t`.
- `ca_j`: Capacidad de la máquina `j` en unidades de tiempo.
- `req_ij`: Tiempo requerido para producir una unidad del producto `i` en la máquina `j`.
- `cp_ij`: Costo de producción por unidad del producto `i` en la máquina `j`.
- `cc_j`: Costo de configuración de la máquina `j`.
- `st_ij`: Tiempo de configuración requerido para el producto `i` en la máquina `j`.
- `ur_ri`: Cantidad de materia prima `r` necesaria para producir una unidad del producto `i`.
- `InvIp_i`: Inventario inicial del producto `i`.
- `InvIr_r`: Inventario inicial de la materia prima `r`.
- `chp_i`: Costo de almacenamiento del producto `i`.
- `cd_i`: Costo de diferimiento para el producto `i`.
- `sm_jt`: Estado de la máquina `j` en el período `t` (1 si está averiada o en tiempo de inactividad, 0 de lo contrario).
- `pb_ijt`: Cantidad de producto `i` producida en la máquina `j` hasta el momento justo de la avería durante el período `t`.
- `p_ijt`: 1 si el producto `i` ha comenzado su producción en una máquina `j` diferente a la averiada en el período de tiempo t, 0 en caso contrario.

### Variables de Decisión

- `X_ijt`: 1 si el producto `i` se asigna a la máquina `j` en el período `t`, 0 en caso contrario.
- `QP_ijt`: Cantidad de producto `i` a producir en la máquina `j` en el período de tiempo `t`.
- `BO_it`: Cantidad diferida del producto `i` en el período de tiempo `t`.
- `InvFP_it`: Inventario final del producto `i` en el período de tiempo `t`.
- `InvFR_rt`: Inventario final de la materia prima `r` en el período de tiempo `t`.
- `XR_ijt`: 1 si el producto `i` se asigna a la máquina `j` en el período de tiempo `t` y período de `t-1`, 0 en caso contrario.
- `Z_jt`: Variable auxiliar dependiente de `XR_ijt`.
- `STT_jt`: Tiempo total de configuración de la máquina `j` en el período de tiempo `t`.
- `STMAX_jt`: Tiempo de configuración tomado del producto `i` con mayor tiempo entre los asignados en máquina `j` en períodos de tiempo `t-1` y `t`.
- `CFT_jt`: Número total de configuraciones en la máquina `j` en el período de tiempo `t`.


### Función Objetivo

El objetivo es minimizar los costos totales, que incluyen los costos de producción, costos de configuración, costos de almacenamiento de inventario y costos de diferimiento.

### Restricciones
### Inventario de Productos Terminados
El inventario final de cada producto `i` en el período `t` se calcula sumando la cantidad producida en ese período al inventario inicial y restando la demanda, considerando cualquier demanda diferida y producción antes de cualquier avería.

### Inventario de Materias Primas
El inventario final de cada materia prima `r` en el período `t` se actualiza en función del inventario inicial y el consumo de materias primas requeridas para la producción de todos los productos. Se asegura la continuidad en el inventario de materias primas a lo largo de los períodos.

### Cumplimiento de la Demanda
La producción total del producto `i` a lo largo de todos los períodos, incluido el inventario inicial, debe cumplir con la demanda del producto, teniendo en cuenta cualquier producción diferida.

### Capacidad de la Máquina
En los períodos en los que una máquina `j` está operativa, la suma de los tiempos de producción y los tiempos de configuración para los productos en la máquina no debe exceder su capacidad. Cuando una máquina está averiada, no se pueden asignar productos a ella.

### Restricciones de Configuración
El tiempo total de configuración para una máquina `j` en el período `t` se calcula en función de los tiempos de configuración para todos los productos asignados.

### Continuidad de la Asignación de Producción
Estas restricciones indican si un producto `i` es asignado o producido en la máquina `j` en los períodos consecutivos `t-1` y `t`, registrando una producción en periodos consecutivos.

### Tiempos de Configuración Máximos
Estas restricciones aseguran que `STMAX_jt` capture el mayor tiempo de configuración entre los productos que se producen en la máquina `j` durante los períodos de tiempo `t-1` y `t`.

### Número Total de Configuraciones
Estas restricciones determinan el número de configuraciones totales en la máquina `j` en cada período de tiempo `t`, considerando posibles ahorros en configuraciones cuando no se requiere un cambio adicional entre los períodos `t-1` y `t`.


### Uso

Para usar este modelo, sigue los siguientes pasos:

**Programación Inicial**

1. **Preparar Datos**: Define los datos de entrada en el archivo correspondiente.
2. **Ejecutar el Modelo**: Ejecuta el archivo `run_production_schedule.py` para resolver el modelo.
3. **Revisar Resultados**: Los resultados se almacenarán en el archivo `Output`.
4. **Verificar Estado**: Al finalizar la programación inicial, el sistema actualizará el archivo `Status` con el valor "1", indicando que la programación inicial ha sido completada. Esto permite al sistema pasar al proceso de detección de eventos sin detener la ejecución del codigo.
5. **Ejecutar programación inicial nuevamente**: En caso de volver a ejecutar una programación inicial, cambiar valor del archivo`Status` a "0" y repetir los pasos

**Reprogramación**

**Simular Evento: Avería de Máquina**

1. **Definir Evento**: Introduce los parámetros correspondientes para simular una avería de máquina en los siguientes archivos:
   - **`Status_machine.tab`**: Archivo donde se define el estado de la máquina (indicando si está en avería o no).
   - **`InProcess_Products.tab`**: Archivo con los productos que han comenzado su producción y que no se pueden mover a otras máquinas (productos inmovibles).
   - **`Produced.tab`**: Archivo con la cantidad producida hasta el momento de los productos asignados inicialmente a la máquina que ha fallado.

2. **Ejecutar Reprogramación**: El código `run_production_schedule.py` detectará la avería y realizará el cálculo de reprogramación necesario.

3. **Obtener Resultados Actualizados**: Los resultados de la reprogramación se guardarán en el archivo `Output`.

**Simular Evento: Escasez de Materia Prima**

1. **Definir Evento**: Introduce los parámetros correspondientes para simular la escasez de materia prima en los siguientes archivos:
   
   - **`BOM.tab`**: Archivo que contiene la lista de materiales asociada a cada producto. Para simular la escasez, asegúrase de que el producto solo use un tipo de materia prima (por ejemplo, `i=1`, `r=1`).
   - **`MPF.tab`**: Archivo donde se define el nuevo periodo de tiempo en que llegará la materia prima faltante (retrasada) y cantidad de materia prima faltante. El periodo de tiempo `t` debe ser mayor a 1.

2. **Ejecutar Reprogramación**: El código `run_production_schedule.py` detectará la escasez de materia prima y realizará el cálculo de reprogramación necesario.

3. **Obtener Resultados Actualizados**: Los resultados de la reprogramación se guardarán en el archivo `Output`.



### Notas
- Para utilizar el generador, asegúrese de agregar manualmente la capacidad de las máquinas (`Data_Machine.tab`), el inventario de materia prima y la lista de materiales (BOM) con la tasa de consumo a sus archivos correspondientes (`Initial_Inventory_RawMaterial.tab` y `BOM.tab`) después de realizar el cálculo de generación de datos.
- Asegúrese de que los datos ingresados en los archivos `BOM.tab`, `MPF.tab`, `Status_machine.tab`, `InProcess_Products.tab`, y `Produced.tab` sean precisos y actualizados antes de ejecutar la reprogramación para obtener resultados correctos.
- Revise regularmente los archivos de entrada para mantener la coherencia y precisión de los datos utilizados por el modelo.



## Python files

### A_S_R.py
Contiene el modelo MILP

### run_production_schedule.py
Ejecuta el algoritmo para resolver un problema concreto.

### Generator.py 
Contiene el código  de generación de datos aleatorios.

### calc_rms.py 
Contiene el algoritmo para calcular pedidos pendientes debido a la escasez de materia prima.



