import os
import pandas as pd

code = '3_2_1'
folder = 'data/'
def cargar_datos():
    ruta_materia_prima = os.path.join(folder, code, 'input', 'MPF.tab')
    ruta_BOM = os.path.join(folder, code, 'input', 'BOM.tab')

    # Cargar datos de materia prima faltante
    df_materia_prima_faltante = pd.read_csv(ruta_materia_prima, sep='\t')
    # Cargar datos de lista de materia prima desde TAB
    df_BOM = pd.read_csv(ruta_BOM, sep='\t')

    return df_materia_prima_faltante, df_BOM

# Función para calcular la cantidad de pedido pendiente
def calcular_pedido_pendiente(df_materia_prima_faltante, df_BOM):
    pedido_pendiente = {}

    for _, row in df_materia_prima_faltante.iterrows():
        if row['IdPeriod'] != 0:
            mp_faltante = row['IdRawMaterial']
            cantidad_faltante = row['Quantity']

            # Encontrar producto relacionado en el archivo BOM
            BOM_relacion = df_BOM[df_BOM['IdRawMaterial'] == mp_faltante]

            for _, bom_row in BOM_relacion.iterrows():
                i = bom_row['IdProduct']
                cantidad_necesaria = bom_row['UtilRate']

                # Evitar la división por cero
                if cantidad_necesaria != 0:
                    unidades_no_fabricadas = cantidad_faltante / cantidad_necesaria

                    # Determinar el periodo de pedido pendiente(demanda diferida)
                    periodo_llegada = row['IdPeriod']
                    t = periodo_llegada - 1

                    # Almacenar el resultado en el diccionario
                    if (i, t) in pedido_pendiente:
                        pedido_pendiente[(i, t)] += unidades_no_fabricadas
                    else:
                        pedido_pendiente[(i, t)] = unidades_no_fabricadas

    return pedido_pendiente

# Ejecutar el programa
if __name__ == "__main__":

    MPF, BOM = cargar_datos()
    # Calcular la pedidos pendientes a entregar(demanda diferida)
    pedido_pendiente = calcular_pedido_pendiente(MPF, BOM)

    # Mostrar resultados
    print("Variables de Pedido pendiente:")
    for (i, t), valor in pedido_pendiente.items():
        print(f"BO[{i}, {t}] = {valor:.1f} unidades")